﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace EXIFLOG
{
    /// <summary>
    /// The metadata of one image, as decoded by exiftool.exe,
    /// presented as a dictionary from labels to values.
    /// </summary>
    public class ImageMetadata : Dictionary<string, string>
    {
        /// <summary>
        /// Complete dump of the metadata, neatly formatted.
        /// </summary>
        public List<string> ToAllStrings()
        {
            var result = new List<string>(100);
            foreach (string key in this.Keys)
            {
                if (key != "Error" || this[key] != "Unknown file type")
                {
                    result.Add(key.PadLeft(40) + " : " + this[key]);
                }
            }

            // Did we fail to get correct metadata?  If so, try to get it another way.
            // It's an SER, FITS, or XISF file.
            // NOTE: Logic is different here than with shorter formats.
            if (this.ContainsKey("Error") && (this["Error"] == "Unknown file type" || this["Error"] == "File format error"))
            {
                var dir = this["Directory"].Replace('/', System.IO.Path.DirectorySeparatorChar);
                var fn = System.IO.Path.Combine(dir, this["File Name"]);

                var desc = SER.Description(fn);
                // There is no "dump" for SER.
                if (desc == "")
                {
                    desc = FITS.Description(fn);
                    if (desc != "") { desc = FITS.Dump(fn); }
                }
                if (desc == "")
                {
                    desc = XISF.Description(fn);
                    if (desc != "") { desc = XISF.Dump(fn); }
                }

                if (desc != "")
                {
                    var lines = desc.Split('\n');
                    result.Add("Header data".PadLeft(40) + " : " + lines[0]);

                    for (int i = 1; i < lines.Length; i++)
                    {
                        result.Add(" ".PadLeft(40) + "   " + lines[i]);
                    }

                }
            }
        
            return result;
        }

        /// <summary>
        /// Selective dump of the metadata, neatly formatted.
        /// </summary>
        public List<string> ToSelectedStrings()
        {
            var result = new List<string>(100);
            foreach (string key in SelectedKeys)
            {
                string val;
                this.TryGetValue(key, out val);
                if (!String.IsNullOrEmpty(val)) result.Add(key.PadLeft(40) + " : " + val);
            }

            if (this.ContainsKey("Error") && (this["Error"] == "Unknown file type" || this["Error"] == "File format error"))
            {
                var dir = this["Directory"].Replace('/', System.IO.Path.DirectorySeparatorChar);
                var fn = System.IO.Path.Combine(dir, this["File Name"]);
                var desc = SER.Description(fn);
                if (desc == "")
                {
                    desc = FITS.Description(fn);
                }
                if (desc == "")
                {
                    desc = XISF.Description(fn);
                }
                if (desc != "")
                {
                    result.Add("Header data".PadLeft(40) + " : " + desc);
                }
            }

            return result;
        }

        public static string[] SelectedKeys =
        {
            "File Name",
            "File Modification Date/Time",
            "Create Date",
            "Camera Model Name",
            "Model",              // alternative to Camera Model Name
            "Image Size",

            "GPS Latitude",
            "GPS Longitude",
            "GPS Altitude",

            "Video Frame Count",   // was Frame Count
            "Frame Rate",
            "Video Codec",
            "Compressor ID",
            "Bit Depth",

            "Exposure Time",
            "F Number",
            "ISO",
            "Focal Length",
            "Lens ID",
            "Lens Spec",

            "Long Exposure Noise Reduction",
            "Long Exposure Noise Reduction 2",
            "High ISO Noise Reduction",
            "Noise Reduction",

            // For FITS as decoded by ExifTool
            "Bitpix",
            "Naxis 1",
            "Naxis 2",
            "Object",
            "Observation Date",
            "Camera",
            "Ccdtemp",
            "Exptime"
        };

        /// <summary>
        /// A one-line summary, suitable for logging.
        /// </summary>
        public string ToString(bool fileNameFirst)
        {
            var sb = new StringBuilder();

            // All files have the first 2 fields, even if they are not images.

            // File name - optional position
            if (fileNameFirst)
            {
                sb.Append(this["File Name"].PadRight(40));
                sb.Append(" ");
            }

            // Date/time, with first 2 colons replaced by hyphens
            string dt = "";
            if (this.ContainsKey("Create Date"))
            {
                dt = this["Create Date"].Substring(0,19);
            }
            else
            {
                dt = this["File Modification Date/Time"].Substring(0,19);
            }
            var dtc = dt.ToCharArray();
            dtc[4] = dtc[7] = '-';
            sb.Append(new String(dtc) + "  ");

            // File name
            if (!fileNameFirst)
            {
                sb.Append(this["File Name"]);
                sb.Append(" : ");  // Blanks ahead of this will be added by the AlignFileNames step.
            }

            // Note: Logic is different from the above 2 cases.
            // If we can parse it as SER, FITS, or XISF using our own decoder, then do so, and bail.
            var dir = this["Directory"].Replace('/', System.IO.Path.DirectorySeparatorChar);
            var fn = System.IO.Path.Combine(dir, this["File Name"]);
            var desc = SER.Description(fn);
            if (desc == "")
            {
                desc = FITS.Description(fn);
            }
            if (desc == "")
            {
                desc = XISF.Description(fn);
            }
            if (desc != "")
            {
                sb.Append(desc);
                return sb.ToString(); // do not attempt to use the other fields
            }

            // Exposure information
            if (this.ContainsKey("Exposure Time")) sb.Append(" " + (this["Exposure Time"] + "s").PadLeft(8));
            if (this.ContainsKey("ISO")) sb.Append(" ISO " + this["ISO"].PadLeft(5));

            // Focal length and f-number if known
            if (this.ContainsKey("F Number") && !this["F Number"].StartsWith("0"))
            {
                if (this.ContainsKey("Focal Length"))
                {
                    sb.Append(this["Focal Length"].Split(' ')[0].PadLeft(7) + "mm");
                }
                sb.Append(" f/" + this["F Number"].PadRight(4));
            }

            // Movie information if present
            if (this.ContainsKey("Video Frame Count")) sb.Append(this["Video Frame Count"].PadLeft(7) + " frames");
            if (this.ContainsKey("Frame Rate")) sb.Append(this["Frame Rate"].PadLeft(7) + " fps");

            string codec = "";
            if (this.ContainsKey("Video Codec"))
            {
                codec = this["Video Codec"];
                if (String.IsNullOrWhiteSpace(codec)) codec = "?";
            }
            else if (this.ContainsKey("Compressor ID"))
            {
                codec = this["Compressor ID"];
                if (String.IsNullOrWhiteSpace(codec)) codec = "?";
            }

            if (codec != "")
            {
                sb.Append("  Codec:" + codec.PadRight(4));
            }

            if (this.ContainsKey("Bit Depth")) sb.Append("  Bits:" + this["Bit Depth"].ToString().PadRight(2));

            // Temperature

            if (this.ContainsKey("Camera Temperature")) sb.Append("  Temp:" + this["Camera Temperature"].Replace(" ","").PadLeft(4));

            // Noise reduction

            string nr = "";
            if (this.ContainsKey("Long Exposure Noise Reduction"))
            {
                nr = this["Long Exposure Noise Reduction"];
            }
            if (nr == "Off" && this.ContainsKey("Long Exposure Noise Reduction 2") && this["Long Exposure Noise Reduction 2"] != "Off")
            {
                nr = this["Long Exposure Noise Reduction 2"];
            }
            if (nr == "" && this.ContainsKey("Noise Reduction"))
            {
                nr = this["Noise Reduction"];
            }
            if (nr != "" && nr != "Off") sb.Append(" NR:" + nr);

            // Live View alias EFCS

            string ef = "";
            if (this.ContainsKey("Live View Shooting"))
            {
                ef = this["Live View Shooting"];
            }
            if (ef == "" && this.ContainsKey("Electronic Front-Curtain Shutter")) // not sure this and the following are actually used
            {
                ef = this["Electronic Front-Curtain Shutter"];
            }
            if (ef == "" && this.ContainsKey("Electronic Front Curtain Shutter")) // not sure this and the following are actually used
            {
                ef = this["Electronic Front Curtain Shutter"];
            }
            if (ef == "" && this.ContainsKey("Electronic Front Curtain")) // not sure this and the following are actually used
            {
                ef = this["Electronic Front Curtain"];
            }
            if (ef != "" && ef != "Off") sb.Append(" EFCS:" + ef);

            // Image size

            if (this.ContainsKey("Image Size")) sb.Append("  " + this["Image Size"].Replace("x"," x ").PadLeft(11));

            // Camera

            string model = "";

            if (this.ContainsKey("Camera Model Name"))
            {
                model = this["Camera Model Name"];
            }
            else if (this.ContainsKey("Model"))
            {
                model = this["Model"];
            }

            string make = "";

            if (this.ContainsKey("Make"))
            {
                make = this["Make"].Split(' ')[0];    // don't go past first blank
            }

            if (!model.StartsWith(make, StringComparison.OrdinalIgnoreCase))
            {
                model = make + " " + model;
            }

            sb.Append("  " + model);

            //// Often incorrect
            //if (this.ContainsKey("Lens ID"))
            //{
            //    sb.Append("  " + this["Lens ID"]);
            //}

            // GPS information if present
            if (this.ContainsKey("GPS Latitude") && this.ContainsKey("GPS Longitude"))
            {
                string[] lat = this["GPS Latitude"].Split(' ');
                string[] lon = this["GPS Longitude"].Split(' ');
                sb.Append("  GPS:");
                sb.Append(lat[4] + lat[0] + "°" + lat[2] + lat[3]);   // Degree sign will be UTF-8
                sb.Append(lon[4] + lon[0] + "°" + lon[2] + lon[3]);
            }


            return sb.ToString();
        }

    }
}
