﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using System.IO;

namespace EXIFLOG
{
    /// <summary>
    /// Decoding of SER file headers.
    /// </summary>
    static class SER
    {
        static Dictionary<int, string> SERColorFormat = new Dictionary<int, string>()
        {
            { 0, "MONO" },
            { 8, "RGGB" },
            { 9, "GRBG" },
            { 10, "GBRG" },
            { 11, "BGGR" },
            { 16, "CYYM" },
            { 17, "YCMY" },
            { 18, "YMCY" },
            { 19, "MYYC" },
            { 100, "RGB" },
            { 101, "BGR" }
        };

        /// <summary>
        /// Decodes a SER file, or returns false if the file is not SER.
        /// </summary>
        /// <param name="filename">Name of file to decode.</param>
        /// <param name="colorFormat">Color format, such as MONO, GRBG, or RGB.</param>
        /// <param name="width">Image width.</param>
        /// <param name="height">Image height.</param>
        /// <param name="bits">Image bits per pixel (e.g., 24 for 8-bit color).</param>
        /// <param name="frames">Number of frames.</param>
        /// <param name="camera">Camera description or other comments inserted by software.</param>
        /// <param name="observer">Observer name or other comments inserted by software.</param>
        /// <param name="telescope">Telescope description or other comments inserted by software.</param>
        /// <param name="timestamp">Time stamp, should be local time but likely to be UTC.</param>
        /// <param name="timestampUTC">Time stamp, UTC.</param>
        public static bool Decode(
            string filename,
            out string colorFormat,
            out int width,
            out int height,
            out int bits,
            out int frames,
            out string camera,
            out string observer,
            out string telescope,
            out DateTime timestamp,
            out DateTime timestampUTC
        )
        {
            width = height = bits = frames = 0;
            camera = observer = telescope = "";
            timestamp = timestampUTC = DateTime.MinValue;
            colorFormat = "?";

            if ((new System.IO.FileInfo(filename)).Length < 178) { return false; }

            using (var br = new BinaryReader(new FileStream(filename, FileMode.Open), Encoding.ASCII))
            {
                var fileID = br.ReadChars(14);
                if (new String(fileID) != "LUCAM-RECORDER") { return false; }

                br.ReadInt32();  // unused

                var colorID = br.ReadInt32();
                if (SERColorFormat.ContainsKey(colorID))
                {
                    colorFormat = SERColorFormat[colorID];
                }

                br.ReadInt32();  // unused

                width = br.ReadInt32();
                height = br.ReadInt32();

                bits = br.ReadInt32();
                if (colorID >= 100) bits = bits * 3;

                frames = br.ReadInt32();

                observer = (new String(br.ReadChars(40))).Replace("\u0000", " ").Trim();
                camera = (new String(br.ReadChars(40))).Replace("\u0000", " ").Trim();
                telescope = (new String(br.ReadChars(40))).Replace("\u0000", " ").Trim();
                // These actually are free-text comments.
                // oaCapture puts nothing there.
                // FireCapture puts a compact mishmash of things.

                timestamp = new DateTime(br.ReadInt64());
                timestampUTC = new DateTime(br.ReadInt64());  // comes out UTC

                return true;
            }
        }

        /// <summary>
        /// If the file can be read as SER, then a string description to be displayed after
        /// the date, time, and filename.  If not SER, then empty string.
        /// </summary>
        public static string Description(string filename)
        {
            if (Decode(filename,
                           out string colorFormat, out int width, out int height, out int bits,
                           out int frames, out string camera, out string observer, out string telescope,
                           out DateTime timestamp, out DateTime timestampUTC))
            {
                // Kluges for FireCapture's unusual format
                camera = camera.Replace("temp=", " temp=");  // kluge for FireCapture ASI
                if (observer.Contains("=")) { camera += " " + Regex.Replace(observer, "([0-9])([a-z])", "$1 $2"); }
                if (telescope.Contains("=")) { camera += " " + Regex.Replace(telescope, "([0-9])([a-z])", "$1 $2"); }

                return
                    String.Format("{0,5} frames {1,4}  {2,4} x {3,-4}  {4,2} bits  {5}",
                                  frames, colorFormat, width, height, bits, camera);
            }
            else
            {
                return "";
            }
        }

    }

    /// <summary>
    /// Decoding of FITS file headers.
    /// </summary>
    static class FITS
    {
        /// <summary>
        /// Decodes a FITS file header.  Note that even the numbers are returned as strings, as found.
        /// If a compressed FITS file is found, all the fields are blank except for a warning in "camera".
        /// </summary>
        /// <returns></returns>
        public static bool Decode(string filename,
            out string width,
            out string height,
            out string depth,
            out string bits,
            out string colorFormat,
            out string camera,
            out string exposure,
            out string iso,
            out string temp
        )
        {
            width = height = bits = camera = colorFormat = exposure = iso = temp = "";
            depth = "1";

            using (var br = new BinaryReader(new FileStream(filename, FileMode.Open), Encoding.ASCII))
            {
                var fileID = new string(br.ReadChars(80));

                if (!fileID.StartsWith("SIMPLE  ="))
                {
                    if (fileID.StartsWith("SIMPLE"))
                    {
                        camera = "(MaxIm DL proprietary compressed FITS format)";
                    }
                    return true;
                }

                // Loop through the header fields

                string field = "";
                while (!field.StartsWith("END "))
                {
                    field = new string(br.ReadChars(80));
                    if (field[8] != '=') continue;  // COMMENT, HISTORY, and the like

                    if (field.StartsWith("NAXIS1")) { width = ValueFITS(field); }
                    else if (field.StartsWith("NAXIS2")) { height = ValueFITS(field); }
                    else if (field.StartsWith("NAXIS3")) { depth = ValueFITS(field); }
                    else if (field.StartsWith("BITPIX")) { bits = ValueFITS(field); }
                    else if (field.StartsWith("COLORSPC")) { colorFormat = ValueFITS(field); }
                    else if (field.StartsWith("INSTRUME")) { camera = ValueFITS(field); }
                    else if (field.StartsWith("CAMERA")) { camera = ValueFITS(field); }
                    else if (field.StartsWith("EXPTIME")) { exposure = ValueFITS(field); }
                    else if (field.StartsWith("ISOSPEED")) { iso = ValueFITS(field); }
                    else if (field.StartsWith("CCDTEMP")) { temp = ValueFITS(field); }
                    else if (field.StartsWith("CCD-TEMP")) { temp = ValueFITS(field); }
                    
                    iso = iso.Replace("ISO", "").Trim();
                }
            }

            return true;
        }

        static string ValueFITS(string field)
        {
            var content = field.Substring(9);  // start after the = sign

            // A value in quotes
            var match = Regex.Match(content, @"^\s'(.+)'");
            if (match.Success)
            {
                return match.Groups[1].Value.Trim();
            };

            // A value not in quotes -- first sequence of nonblanks in the content
            match = Regex.Match(content, @"^\s*(\S+)");
            if (match.Success)
            {
                var result = match.Groups[1].Value;

                // If it parses as a number, parse it and reformat it with limited significant digits
                double x;
                if (double.TryParse(result, out x))
                {
                    result = x.ToString("#.####");
                }

                return result;
            };

            // Failed to find a value
            return "";
        }

        /// <summary>
        /// The entire FITS header, if found, with each item on a new line.
        /// Otherwise empty string.  File is presumed to be known to be FITS
        /// (e.g., Decode has been run on it successfully).
        /// </summary>
        public static string Dump(string filename)
        {
            var sb = new StringBuilder();

            using (var br = new BinaryReader(new FileStream(filename, FileMode.Open), Encoding.ASCII))
            {
                var fileID = new string(br.ReadChars(80));

                if (!fileID.StartsWith("SIMPLE  ="))
                {
                    if (fileID.StartsWith("SIMPLE"))
                    {
                        return "MaxIm DL proprietary compressed FITS format.";
                    }
                    return "";
                }

                // Loop through the header fields

                string field = "";
                while (!field.StartsWith("END "))
                {
                    field = new string(br.ReadChars(80));
                    sb.AppendLine(field);
                }

                return sb.ToString();
            }
        }

        /// <summary>
        /// If the file can be read as FITS, then a string description to be displayed after
        /// the date, time, and filename.  If not FITS, then empty string.
        /// </summary>
        public static string Description(string filename)
        {
            try
            {
                if (Decode(filename,
                    out string width, out string height, out string depth, out string bits, out string colorFormat,
                    out string camera, out string exposure, out string iso, out string temp)
                )
                {
                    if (width == "")
                    {
                        return camera;  // message about compressed FITS appears here
                    }

                    string result = exposure.PadLeft(5) + "s";
                    if (iso != "") { result += ("ISO:" + iso).PadLeft(10); }
                    result += (width + " x " + height + " x " + depth).PadLeft(20) + "  " + bits + " bits " + camera;
                    if (colorFormat != "") { result += " " + colorFormat; }
                    if (temp != "") { result += " Temp:" + temp + "C"; }

                    return result;
                }
                else
                {
                    return "";
                }
            }
            catch
            {
                return "";
            }
        }
    }

    /// <summary>
    /// Partial decoding of XISF files.  Parallels the FITS class closely.
    /// </summary>
    static class XISF
    {
        /// <summary>
        /// Decodes a XISF file header.  Note that even the numbers are returned as strings, as found.
        /// </summary>
        /// <returns></returns>
        public static bool Decode(string filename,
            out string width,
            out string height,
            out string depth,
            out string bits,
            out string colorFormat,
            out string camera,
            out string exposure,
            out string iso,
            out string temp
        )
        {
            width = height = depth = bits = camera = colorFormat = exposure = iso = temp = "";

            using (var br = new BinaryReader(new FileStream(filename, FileMode.Open), Encoding.ASCII))
            {
                var xisfIdentifier = new string(br.ReadChars(8));
                if (!xisfIdentifier.StartsWith("XISF")) return false;

                var headerLength = br.ReadUInt32();
                br.ReadBytes(4);  // reserved field
                var headerText = new string(br.ReadChars((int)headerLength));

                // We could parse the XML.  However, for now, we will simply find things using regexes.

                var match = Regex.Match(headerText, "<Image.+?geometry=\"(.+?)\"");
                if (match.Success)
                {
                    var geometry = match.Groups[1].Value;
                    var fields = geometry.Split(':');
                    width = fields[0];
                    height = fields[1];
                    depth = fields[2];
                }

                match = Regex.Match(headerText, "<Image.+?sampleFormat=\"(.+?)\"");
                if (match.Success) { bits = match.Groups[1].Value.Replace("UInt", ""); }

                match = Regex.Match(headerText, "<Image.+?colorSpace=\"(.+?)\"");
                if (match.Success) { colorFormat = match.Groups[1].Value; }

                match = Regex.Match(headerText, "<Property id=\"Instrument:ExposureTime\".*?value=\"(.+?)\"");
                if (match.Success) { exposure = match.Groups[1].Value; }

                match = Regex.Match(headerText, "<Property id=\"Instrument:Camera:ISOSpeed\".*?value=\"(.+?)\"");
                if (match.Success) { iso = match.Groups[1].Value; }



                match = Regex.Match(headerText, "<Property id=\"Instrument:Camera:Name\".*?>(.+?)<");
                if (match.Success) { camera = match.Groups[1].Value; }

                // Embedded FITS fields.  If present, these supersede.

                match = Regex.Match(headerText, "<FITSKeyword name=\"EXPTIME\".*?value=\"(.+?)\"");
                if (match.Success) { exposure = Dequote(match.Groups[1].Value); }

                match = Regex.Match(headerText, "<FITSKeyword name=\"INSTRUME\".*?value=\"(.+?)\"");
                if (match.Success) { camera = Dequote(match.Groups[1].Value); }

                match = Regex.Match(headerText, "<FITSKeyword name=\"CAMERA\".*?value=\"(.+?)\"");
                if (match.Success) { camera = Dequote(match.Groups[1].Value); }

                match = Regex.Match(headerText, "<FITSKeyword name=\"BITPIX\".*?value=\"(.+?)\"");
                if (match.Success) { bits = Dequote(match.Groups[1].Value); }

                match = Regex.Match(headerText, "<FITSKeyword name=\"CCD\\-TEMP\".*?value=\"(.+?)\"");
                if (match.Success) { temp = Dequote(match.Groups[1].Value); }

                match = Regex.Match(headerText, "<FITSKeyword name=\"CCDTEMP\".*?value=\"(.+?)\"");
                if (match.Success) { temp = Dequote(match.Groups[1].Value); }

            }

            return true;
        }

        private static string Dequote(string s)
        {
            if (s.StartsWith("'") && s.EndsWith("'"))
            {
                return s.Substring(1, s.Length - 2).Trim();
            }
            else
            {
                return s;
            }
        }

        /// <summary>
        /// The entire XISF header, if found, with each item on a new line.
        /// Otherwise empty string.  File is presumed to be known to be XISF
        /// (e.g., Decode has been run on it successfully).
        /// </summary>
        public static string Dump(string filename)
        {
            using (var br = new BinaryReader(new FileStream(filename, FileMode.Open), Encoding.ASCII))
            {
                var xisfIdentifier = new string(br.ReadChars(8));
                if (!xisfIdentifier.StartsWith("XISF")) return "";

                var headerLength = br.ReadUInt32();
                br.ReadBytes(4);  // reserved field
                var headerText = new string(br.ReadChars((int)headerLength));

                return headerText.Replace("<", "\n<").Substring(1);
            }
        }

        /// <summary>
        /// If the file can be read as XISF, then a string description to be displayed after
        /// the date, time, and filename.  If not XISF, then empty string.
        /// </summary>
        public static string Description(string filename)
        {
            // This is just like FITS.Description except for changes actually needed.

            try
            {
                if (Decode(filename,
                    out string width, out string height, out string depth, out string bits, out string colorFormat,
                    out string camera, out string exposure, out string iso, out string temp)
                )
                {
                    if (width == "")
                    {
                        return camera;  // message about compressed FITS appears here
                    }

                    string result = exposure.PadLeft(5) + "s";
                    if (iso != "") { result += ("ISO:" + iso).PadLeft(10); }
                    result += (width + " x " + height + " x " + depth).PadLeft(20) + "  " + bits + " bits " + camera;
                    if (colorFormat != "") { result += " " + colorFormat; }
                    if (temp != "") { result += " Temp:" + temp + "C"; }
                    // Presently temp is never decoded from XISF

                    return result;
                }
                else
                {
                    return "";
                }
            }
            catch
            {
                return "";
            }
        }
    }

}
